
#include "poppet.h"
#include "resources.h"

#include <limits.h>
#include <stdlib.h>
#include <string.h>
#include <dirent.h>
#include <fcntl.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/stat.h>
#include <errno.h>
#include <time.h>

/* message dispositions */
#define D_SEEN		(1<<0)
#define D_ANSWERED	(1<<1)
#define D_DELETED	(1<<2)
#define D_FLAGGED	(1<<3)
#define D_RECENT	(1<<4)
#define D_DRAFT		(1<<5)


#define MAILDIR_CHAR '-'
static unsigned int MaildirCount = 0;



int
maildir_open (const char *path)
{
    char buf[_POSIX_PATH_MAX];
    struct stat sb;
    const char *subdirs[] = { "cur", "new", "tmp" };
    int i;

    if (stat (path, &sb))
    {
	if (errno == ENOENT)
	{
	    if (mkdir (path))
	    {
		log_printf ("ERROR: mkdir %s: %s (errno %d)\n",
			 path, strerror (errno), errno);
		goto err;
	    }

	    for (i = 0; i < 3; i++)
	    {
		snprintf (buf, sizeof (buf), "%s/%s", path, subdirs[i]);
		if (mkdir (buf))
		{
		    log_printf ("ERROR: mkdir %s: %s (errno %d)\n",
			     buf, strerror (errno), errno);
		    goto err;
		}
	    }

	}
	else
	{
	    log_printf ("ERROR: stat %s: %s (errno %d)\n", path,
		     strerror (errno), errno);
	    goto err;
	}
    }
    else
    {
	/* check to make sure this looks like a valid maildir box */
	for (i = 0; i < 3; i++)
	{
	    snprintf (buf, sizeof (buf), "%s/%s", path, subdirs[i]);
	    if (stat (buf, &sb))
	    {
		log_printf ("ERROR: stat %s: %s (errno %d)\n", buf,
			 strerror (errno), errno);
		log_printf ("ERROR: %s does not appear to be a valid maildir style mailbox\n",
			 path);
		goto err;
	    }
	}
    }

    return 0;

  err:
    return -1;
}


#define _24_HOURS (3600 * 24)

static void
maildir_clean_tmp (const char *mbox)
{
    char path[_POSIX_PATH_MAX];
    DIR *dirp;
    struct dirent *entry;
    struct stat info;
    time_t now;

    snprintf (path, sizeof (path), "%s/tmp", mbox);
    dirp = opendir (path);
    if (dirp == NULL)
    {
	log_printf ("maildir_clean_tmp: opendir: %s: %s (errno %d)\n",
		 path, strerror (errno), errno);
	return;
    }
    /* assuming this scan will take less than a second, we only need to
     * check the time once before the following loop.
     */
    time (&now);
    while ((entry = readdir (dirp)))
    {
	snprintf (path, sizeof (path), "%s/tmp/%s", mbox, entry->d_name);
	if (stat (path, &info))
	    log_printf ("maildir_clean_tmp: stat: %s: %s (errno %d)\n",
		     path, strerror (errno), errno);
	else if (S_ISREG (info.st_mode) && now - info.st_ctime >= _24_HOURS)
	{
	    /* this should happen infrequently enough that it won't be
	     * bothersome to the user to display when it occurs.
	     */
	    log_printf ("Warning: removing stale file %s\n", path);
	    if (unlink (path))
		log_printf (
			 "maildir_clean_tmp: unlink: %s: %s (errno %d)\n",
			 path, strerror (errno), errno);
	}
    }
}


int
maildir_close (const char *path)
{
    /* per the maildir(5) specification, delivery agents are supposed to
     * set a 24-hour timer on items placed in the `tmp' directory.
     */
    maildir_clean_tmp (path);

    return 0;
}



int maildir_save (char *path, char *src, int fsize, int lines)
{
  int ret, len, flags = 0;
  char dest[_POSIX_PATH_MAX];
  char suffix[_POSIX_PATH_MAX];
  char newpath[_POSIX_PATH_MAX];
  int fd;

  len = strlen(path);
  if (len > 0 && (path[len-1]=='/' || path[len-1]=='\\'))
    path[len-1] = 0;

  if (maildir_open(path) != 0)
    return -1;

  /* construct the flags part of the file name. */
  
  *suffix = 0;
  if (flags & ~D_RECENT)
  {
    snprintf (suffix, sizeof (suffix), "%c2,%s%s%s%s", MAILDIR_CHAR,
	      (flags & D_FLAGGED) ? "F" : "",
	      (flags & D_ANSWERED) ? "R" : "",
	      (flags & D_SEEN) ? "S" : "",
	      (flags & D_DELETED) ? "T" : "");
  }
  
  /* create new file */
  snprintf (dest, sizeof (dest), "%s/tmp/%ld_%d.%d.%s%s",
	    path, time (0), MaildirCount++, getpid (),
	    Hostname, suffix);
  
  if ((fd = open (dest, O_WRONLY | O_CREAT | O_EXCL | O_BINARY, 0600)) <= 0)
  {
    if (errno != EEXIST)
    {
      perror (dest);
    }
    return -1;
  }
  close(fd);

  ret = file_cat(dest, src, fsize, lines);

  if (!ret) {
    char *p = strrchr (dest, '/');

    snprintf (newpath, sizeof (newpath), "%s/%s%s", path,
	      (flags & ~D_RECENT) ? "cur" : "new", p);

    if ( rename (dest, newpath) ) {
      perror ("rename");
      log_printf("rename (%s, %s)\n", dest, newpath);
      ret = -1;
    }
  }
  /* always remove the temp file */
  if (!ret)
    unlink (dest);

  maildir_close(path);

  return ret;
}
